<?php

/**
 * @file
 * Theme preprocessor functions.
 */
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Element;

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet(&$vars) {
  $vars['canonical'] = $vars['element']['#mcapi_wallet']->toUrl('canonical')->toString();
  $vars['attributes']['class'] = ['wallet', $vars['element']['#view_mode']];
}

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet_component(&$vars) {
  $vars['title'] = @$vars['element']['#title'];
  $vars['attributes']['class'][] = 'component';
  $vars['attributes']['class'][] = str_replace('_', '-', substr($vars['element']['#theme'], 7));
}

/**
 * Theme preprocessor for all of an entity's wallets.
 */
function template_preprocess_mcapi_wallets(&$vars) {
  $viewbuilder = \Drupal::entityTypeManager()->getViewBuilder('mcapi_wallet');
  // Convert each wallet into a render array.
  foreach ($vars['wallets'] as $key => $wallet) {
    $vars['wallets'][$key] = $viewbuilder->view($wallet, $vars['view_mode']);
  }
}

/**
 * Theme preprocessor for 'balance_bars'.
 *
 * All values sent to the theme system as numbers should be formatted plainly,
 * all labels formatted normally.
 */
function template_preprocess_wallet_balance_bars(&$vars) {
  $wallet = $vars['wallet'];
  $currency = $vars['currency'];
  $stats = $wallet->getStats($currency->id);
  $id = "given-received-" . $currency->id . '-' . $wallet->id();
  $max_raw = max($stats->gross_in, $stats->gross_out);
  $max_formatted_plain = $currency->format(
    _mcapi_get_axis_max($max_raw),
    CurrencyInterface::DISPLAY_PLAIN
  );
  $vars += [
    'label' => $currency->label(),
    'id' => $id,
    'functionname' => str_replace('-', '_', $id),
    'incoming' => $currency->format($stats->gross_in, CurrencyInterface::DISPLAY_PLAIN),
    'outgoing' => $currency->format($stats->gross_out, CurrencyInterface::DISPLAY_PLAIN),
    'show_in' => $currency->format($stats->gross_in, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']),
    'show_out' => $currency->format($stats->gross_out, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']),
    'max' => $max_formatted_plain,
    'vaxislabels' => [
      [
        'value' => 0,
        'label' => strip_tags($currency->format(0, CurrencyInterface::DISPLAY_NORMAL)),
      ],
      [
        'value' => floor($max_formatted_plain / 2),
        'label' => strip_tags($currency->format(intval($max_raw / 2), CurrencyInterface::DISPLAY_NORMAL)),
      ],
      [
        'value' => $max_formatted_plain,
        'label' => strip_tags($currency->format($max_raw, CurrencyInterface::DISPLAY_NORMAL)),
      ],
    ],
  ];
}

/**
 * Template preprocessor for 'wallet_balance'.
 */
function template_preprocess_wallet_balance_viz(&$vars) {
  $stats = $vars['wallet']->getStats($vars['currency']->id);
  if ($stats->gross_in > $stats->gross_out) {
    $vars['side'] = 'positive';
  }
  elseif ($stats->gross_in < $stats->gross_out) {
    $vars['side'] = 'negative';
  }
  else {
    $vars['side'] = 'balanced';
  }
  $vars['income'] = $vars['currency']->format($stats->gross_in, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
  $vars['expenditure'] = $vars['currency']->format($stats->gross_out, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
  $vars['balance'] = $vars['currency']->format($stats->balance, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
}

/**
 * Theme preprocessor for 'wallet_balance'.
 * Should only be available when there is more than one currency.
 */
function template_preprocess_wallet_balance(&$vars) {
  $bal = $vars['wallet']->getStats($vars['currency']->id())->balance;
  $vars['worth'] = Markup::Create($vars['currency']->format($bal, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']));
}

/**
 * Theme preprocessor for 'wallet_stats'.
 */
function template_preprocess_wallet_stats(&$vars) {
  $unused = \Drupal::config('mcapi.settings')->get('render_unused');
  foreach ($vars['wallet']->getAllStats($unused) as $curr_id => $data) {
    $currency = Currency::load($curr_id);
    $vars['currencies'][$curr_id]['label'] = $currency->label();
    $vars['currencies'][$curr_id]['gross_in'] = $currency->format($data->gross_in, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
    $vars['currencies'][$curr_id]['gross_out'] = $currency->format($data->gross_out, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
    $vars['currencies'][$curr_id]['volume'] = $currency->format($data->volume, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
    $vars['currencies'][$curr_id]['balance'] = $currency->format($data->balance, CurrencyInterface::DISPLAY_NORMAL, $vars['metrics_link']);
    $vars['currencies'][$curr_id]['partners'] = $data->partners;
    $vars['currencies'][$curr_id]['trades'] = $data->trades;
  }
}

/**
 * Theme preprocessor for 'mcapi_transaction'.
 */
function template_preprocess_mcapi_transaction(&$vars, $hook, $info) {
  $elements = $vars['elements'];
  foreach (Element::children($vars['elements']) as $key) {
    $vars[$key] = \Drupal::service('renderer')->render($vars['elements'][$key]);
  }
  if (!empty($elements['#watermark'])) {
    // to add a watermark to the render array, add #watermark => blah
    $vars['watermark'] = Markup::create('<div class = "watermark">'.$elements['#watermark'].'</div>');
  }
  if (isset($elements['#links'])) {
    $vars['links'] = $elements['#links'];
  }

  $vars['created'] = \Drupal::service('date.formatter')->format($vars['elements']['#mcapi_transaction']->created->value);
}

/**
 * Theme preprocessor for 'wallet_balance_history'.
 */
function template_preprocess_wallet_balance_history(&$vars) {
  $currency = $vars['currency'];
  $vars['id'] = 'wid_'.$vars['wallet']->id().'_curr_'.$currency->id;
  $points = &$vars['points'];
  // Add a start and end points showing the balance at this moment.
  $points = [$vars['wallet']->created->value => 0] + $points;
  $points[\Drupal::time()->getRequestTime()] = end($points);

  //$points[1718445004] *= -1;
  // Apply smoothing, or even roughing.
  if (count($points) < $vars['width'] / 3) {
    // Step method, for a small number of points.
    $times = $values = [];
    // Make two values for each one in the keys and values.
    foreach ($points as $time => $bal) {
      $times[] = $time;
      $times[] = $time + 1;
      $values[] = $bal;
      $values[] = $bal;
    }
    // Now slide the arrays against each other to create steps.
    array_pop($values);
    array_shift($times);
    $points = array_combine($times, $values);
  }
  elseif (count($points) > $vars['width']) {
    // Decimate the array, for a large number of points.
    $ratio = count($points) / $vars['width'];
    $factor = intval($ratio + 1);
    // Now iterate through the array taking 1 out of every $factor values.
    $i = 0;
    foreach ($points as $key => $value) {
      if ($i % $factor != 0) {
        unset($points[$key]);
      }
      $i++;
    }
  }
  // Prepare for javascript
  foreach ($points as $timestamp => $val) {
    $vars['daterows'][date('Y,m,d', $timestamp)] = $val;
  }
  $vars['vaxislabels'] = gchart_axes($currency, $points);
  if ($vars['format_curr']) {
    $vars['daterows'] = array_map(function($v) use ($currency) {return $currency->format($v, CurrencyInterface::DISPLAY_PLAIN);}, $vars['daterows']);
  }
  $vars['height'] = (int)$vars['width']/2;
  $vars['attributes']['class'] = ['component', 'balance_history'];
}

/**
 * Calculate some good axis labels, using the min & max balance extents.
 *
 * @param array $vals
 *   The $points history of raw balances keyed by unixtimes.
 *
 * @return array
 *   The min, max and mid axis labels and values.
 */
function gchart_axes(CurrencyInterface $currency, $vals) : array {
  $max = max($vals);
  $min = min($vals);
  if ($min >= 0) {
    $max_axis_plain = _mcapi_get_axis_max($currency->format($max, CurrencyInterface::DISPLAY_PLAIN));
    $max_axis_raw = $currency->unformat(['main' => $max_axis_plain]);
    $mid_axis_raw = $currency->unformat(['main' => intval($max_axis_plain/2)]);
    $min_axis_plain = 0;
    $min_axis_formatted = $currency->format(0);
    $max_axis_formatted = $currency->format($max_axis_raw);
  }
  elseif ($max <= 0) {
    $min_axis_plain = -_mcapi_get_axis_max($currency->format(abs($min), CurrencyInterface::DISPLAY_PLAIN));
    $min_axis_raw = $currency->unformat(['main' => $min_axis_plain]);
    $mid_axis_raw = $currency->unformat(['main' => intval($min_axis_plain/2)]);
    $max_axis_plain = 0;
    $min_axis_formatted = $currency->format($min_axis_raw);
    $max_axis_formatted = $currency->format(0);
  }
  else {
    $min_axis_plain = -_mcapi_get_axis_max($currency->format(abs($min), CurrencyInterface::DISPLAY_PLAIN));
    $mid_axis_raw = 0;
    $max_axis_plain = _mcapi_get_axis_max($currency->format($max, CurrencyInterface::DISPLAY_PLAIN));
    $max_axis_formatted = $currency->format(0);
  }

  return [
    [
      'value' => $min_axis_plain,
      'label' => strip_tags($min_axis_formatted)
    ],
    [
      'value' => $currency->format(intval($mid_axis_raw), CurrencyInterface::DISPLAY_PLAIN),
      'label' => strip_tags($currency->format($mid_axis_raw))
    ],
    [
      'value' => $max_axis_plain,
      'label' => strip_tags($max_axis_formatted)
    ]
  ];
}

/**
 * Helper.
 *
 * @param float $val
 *   The plain formatted value to represent in the chart.
 */
function _mcapi_get_axis_max($plain_val) : int {
  $max = 10 ** ceil(log($plain_val, 10));
  while ($max/2 > $plain_val) {
    $max = $max / 2;
  }
  return intval($max);
}

/**
 * Theme preprocessor for 'mcapi_currency'
 */
function template_preprocess_mcapi_currency(&$variables) {
  $currency = $variables['elements']['#mcapi_currency'];
  $stats = \Drupal::entityTypeManager()
    ->getStorage('mcapi_transaction')
    ->ledgerStateQuery($currency->id(), [])
    ->execute()
    ->fetch();
  $variables['name'] = $currency->label();
  $variables['volume'] = $currency->format($stats->volume, CurrencyInterface::DISPLAY_NORMAL, TRUE);
  $variables['trades'] = $stats->trades;
  $variables['wallets'] = $stats->partners;
  $variables['url'] = $currency->toUrl();
}
