<?php

namespace Drupal\mcapi_credcom;

use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi_credcom\CredCom;
use Drupal\mcapi_credcom\FinalExceptionSubscriber;
use Drupal\mcapi_credcom\Functions;
use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\WorkflowViolation;
use CreditCommons\Workflow as CCWorkflowInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\TempStore\SharedTempStore;

/**
 * Handle all incoming requests from the trunkward node.
 */
class EndPoint extends \Drupal\Core\Controller\ControllerBase {

  private $logger;
  private $config;
  private $walletLookup = [];
  private $tempstore;

  /**
   * See HashChecker for more set up operations.
   */
  public function __construct(Request $request, SharedTempStore $credcom_tempstore) {
    global $error_context;
    $this->config = $this->config('mcapi_credcom.settings');
    $this->logger = $this->getLogger('mcapi_credcom');
    $this->tempstore = $credcom_tempstore;
    $error_context = (object)[
      'node' => $this->config->get('cc_user'),
      'method' => $request->getMethod(),
      'path' => $request->getPathInfo(),
      'user' => $this->config->get('trunkward_node_id')
    ];
    \Drupal::getContainer()->set('exception.final', new FinalExceptionSubscriber());
  }

  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('tempstore.shared')->get('mcapi_credcom')
    );
  }

  /**
   * Having validated with the hash, return that this leaf node is online
   */
  function handshake() {
    return JsonResponse::create([]);
  }

  /**
   * Return the exchange rate.
   */
  function convert() {
    $result = [
      'data' => [
        'symbol' => $this->config->get('local_curr_symbol'),
        'rate' => 1 * $this->config->get('rate')
      ]
    ];
    return JsonResponse::create($result);
  }

  /**
   * @param Request $request
   * @return JsonResponse
   */
  function autocompleteNames(Request $request) {
    $fragment = $request->query->get('acc_path');
    $wids = \Drupal::entityQuery('mcapi_wallet')
      ->condition('name', '%'.$fragment.'%', 'LIKE')
      ->range(0, 10)
      ->execute();
    $result = [];
    foreach (Wallet::loadMultiple($wids) as $wid => $wallet) {
      // Credit commons wallet names can't have spaces.
      $result[] = str_replace(' ', '-', $wallet->label());
    }
    return JsonResponse::create($result);
  }

  /**
   * Handle incoming transaction request.
   * @return JsonResponse
   */
  function relayTransaction(Request $request) {
    $contents = $request->getContent();
    $this->logger->debug('Received json @credcom', ['@credcom' => $contents]);
    $credComFull = json_decode($contents);
    foreach ($credComFull->entries as &$entry) {
      $entry = Functions::upcastEntry($entry);
    }
    $credCom = new CredCom($credComFull->uuid, $credComFull->type, $credComFull->entries);
    $transaction = Functions::convertCredComToDrupal($credCom);
    $transaction->credCom = $credCom;
    if ($credCom->workflow()->creation->confirm) {
      $credCom->changeState(CCWorkflowInterface::STATE_VALIDATED);
      $this->tempstore->set($credCom->uuid, $transaction);
      $response_code = 200;
    }
    else {
      $transaction->save();
      $response_code = 201;
    }
    return JsonResponse::create($transaction->credCom->additional, $response_code);
  }

  /**
   * For unimplemented CreditCommons REST methods
   */
  function returnArray() {
    return JsonResponse::create([]);
  }


  /**
   *
   * @param string $uuid
   * @param string $state
   * @return JsonResponse
   * @throws WorkflowViolation
   */
  function changeState(string $uuid, string $state) {
    $transaction = $this->tempstore->get($uuid);
    if (!$transaction) {
      $txs = \Drupal::entityTypeManager()->getStorage('mcapi_transaction')->loadByProperties(['uuid' => $uuid]);
      if (empty($txs)) {
        throw new CCFailure("Transaction $uuid not found in Drupal tempStore.");
      }
      $transaction = reset($txs);
    }
    if (!isset($transaction->credCom)) {
      throw new CCFailure("Transaction $uuid seems not to be a remote transaction.");
    }
    $credCom = $transaction->credCom;

    $permitted = $credCom->workflow()->canTransitionToState(
      \Drupal::config('mcapi_credcom.settings')->get('trunkward_node_id'),
      $credCom->state,
      $credCom->entries[0],
      $state
    );

    if (!$permitted) {
      throw new WorkflowViolation(type: $credCom->type, from: $credCom->state, to: $state);
    }

    $credCom->changeState($state);
    $transaction->state->setValue(array_search($state, CredCom::MCAPI_CREDCOM_STATE_MAP));
    if ($state == 'pending') {// Equivalent to mcapi_signatures_mcapi_entity_presave()
      $local_wallet = ($transaction->payee->target_id == \Drupal::config('mcapi_credcom.settings')->get('wallet_id')) ?
        $transaction->payer->entity :
        $transaction->payee->entity;
      $transaction->signatures[$local_wallet->getOwner()->id()] = 0;
    }
    $transaction->save();
    $this->tempstore->delete($uuid);
    return JsonResponse::create(NULL, 201);
  }


}
