<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Transaction as CCTransaction;
use CreditCommons\Exceptions\CCError;
use Drupal\mcapi_credcom\Account;
use Drupal\mcapi_credcom\AccountRemote;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\Core\Render\Markup;

/**
 * Access check for requests from trunkwards nodes.
 */
class Functions {

  static $remoteAccPath;

  /**
   * @param Transaction $transaction
   * @return bool
   *   TRUE if either the payer or payee is the balance of trade wallet.
   */
  static function isRemote(Transaction $transaction) : bool {
    $bot_wid = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    return $transaction->payee->target_id == $bot_wid or $transaction->payer->target_id == $bot_wid;
  }

  /**
   * Convert the given credcom transaction into a Drupal transaction.
   * The given entries are converted to children and any children added by
   * Drupal are converted to Entries and added.
   *
   * @param CredCom $credcom
   * @return Transaction
   * @throws CCFailure
   */
  static function convertCredcomToDrupal(CredCom $credcom) : Transaction {
    $remote_wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    // Convert credcom entries into the transaction and child transactions.
    $props = static::convertCCEntryToDrupalValues($credcom->entries[0]);
    $props['uuid'] = $credcom->uuid;
    $transaction = Transaction::create($props);
    foreach (array_slice($credcom->entries, 1) as $entry) {
      $props = static::convertCCEntryToDrupalValues($entry);
      try {
        $child = $transaction->createChild($props);
      }
      catch (\Exception $e) { }
    }

    foreach ($transaction->assemble() as $child) {
      if (static::isRemote($child)) {
        $e = static::ConvertTransactionToCcEntry($child);
        $credcom->entries[] = $e;
        $credcom->additional[] = $e;
      }
    }
    foreach ($transaction->validate() as $violation) {
      /** @var Symfony\Component\Validator\ConstraintViolation $violation */
      $message = 'Error in new transaction: '.$violation->getMessage();
      \Drupal::logger('mcapi_credcom')->debug('@message', ['@message' => $message]);
      throw new CCFailure($message);
    }
    return $transaction;
  }

  /**
   *
   * @param \stdClass $entry
   * @return array
   * @throws CCFailure
   */
  private static function convertCCEntryToDrupalValues($entry) : array  {
    $config = \Drupal::config('mcapi_credcom.settings');
    $props = [
      'type' => 'remote',
      'curr_id' => $config->get('currency_id'),
      'payee' => self::convertCCaddrToWallet($entry->payee),
      'payer' => self::convertCCaddrToWallet($entry->payer),
      'quantity' => intval($entry->quant / $config->get('rate')),
      'description' => $entry->description
    ];

    // One and only one wallet is the intertrading wallet.
    if (!($props['payee'] == $config->get('wallet_id') xor $props['payer'] == $config->get('wallet_id'))) {
      throw new CCFailure('Cannot make remote transaction with '.$entry->payee.'('.$props['payee'].') and '.$entry->payer .'('.$props['payer'].')');
    }
    return $props;
  }

  /**
   * Get the wallet ID from the credcom path.
   * @param string $cc_path
   *   A Credit Commons account path
   * @return int
   */
  private static function convertCCaddrToWallet(string $cc_path) : int {
    $config = \Drupal::config('mcapi_credcom.settings');
    $local_acc_name = static::credcomPathIsLocal($cc_path);

    if (!$local_acc_name) {
      $wid = $config->get('wallet_id');
    }
    elseif (!isset($wid) and is_numeric($local_acc_name)) {
      // Assuming its not a float!
      $wid = intval($local_acc_name);
    }
    if (!isset($wid)){
      $entityTypeManager = \Drupal::entityTypeManager();
      // Wallet name
      $wids = $entityTypeManager->getStorage('mcapi_wallet')
        ->getQuery()->condition('name', $local_acc_name)
        ->accessCheck(FALSE)
        ->execute();
      if ($wids) {
        $wid = reset($wids);
      }
    }
    if (!isset($wid)) {
      // User email address
      if (filter_var($local_acc_name, FILTER_VALIDATE_EMAIL)) {
        $users = $entityTypeManager->getStorage('user')->loadByProperties(['mail' => $local_acc_name]);
        $user = reset($users);
      }
      // User name
      else {
        $uids = $entityTypeManager->getStorage('user')
          ->getQuery()->condition('name', $local_acc_name)
          ->accessCheck(FALSE)
          ->execute();
        if ($uids)  {
          $user = \Drupal\user\Entity\User::load(reset($uids));
        }
      }
      if (isset($user)) {
        $wallet = WalletStorage::firstWalletOfEntity($user);
        if ($wallet) {
          $wid = $wallet->id();
        }
      }
    }
    if (isset($wid)) {
      return $wid;
    }
    throw new \CreditCommons\Exceptions\DoesNotExistViolation(id: $cc_path, type: 'account');
  }

  /**
   * Entity builder callback.
   *
   * Replace the remote wallet with the intertrading wallet and build the credcom
   * pseudo property.
   */
  static function entityBuilder($entity_type_id, $entity, $form, $form_state) {
    if ($entity_type_id == 'mcapi_transaction' and $entity->type->target_id == 'remote') {
      $remote_wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
      // the remote wallet still has a value of the account path.
      if (isset($entity->payer->target_id)) {// outgoing
        $entity->payee->target_id = $remote_wallet_id;
        static::$remoteAccPath = $entity->payee->value;
      }
      else  {
        $entity->payer->target_id = $remote_wallet_id;
        static::$remoteAccPath = $entity->payer->value;
      }
      // Drupal builds the entity in both the validation and submit phases of the form.
      // We save the remote result to save doing the request twice.
      if (!isset($form_state->credCom)) {
        $credcom = [
          'uuid' => $entity->uuid(),
          'type' => substr($form_state->get('form_display')->getMode(), strlen('remote_')),
          'entries' => [static::ConvertTransactionValuesToCcEntry($form_state->getValues())],
        ];
        $cc_transaction = CCTransaction::create((object)$credcom);

        try {
          // remote build and validation
          $new_rows = \Drupal::service('mcapi_credcom.trunkward_node')->buildValidateRelayTransaction($cc_transaction);
        }
        catch (\CreditCommons\Exceptions\CCError $e) {
          $form_state->setErrorByName('', $e->makeMessage() .' '. t('See error log for details'));
          \Drupal::logger('mcapi_credcom')->warning(
            'Failed to build transaction: @error',
            ['@error' => Markup::create('<pre>'.print_r($e, 1).'</pre>')]
          );
        }
        foreach ($new_rows as $row) {
          $credcom['entries'][] = static::upcastEntry($row);
          $props = Functions::convertCCEntryToDrupalValues($row);
          try {
            $child = $entity->createChild($props);
          }
          catch (\Exception $e) { }
        }
        $form_state->credCom = new CredCom(...$credcom);
        $form_state->credCom->changeState('validated');
      }
      $entity->credCom = $form_state->credCom;
    }

    return $entity;
  }

  /**
   * Entity form validation callback.
   *
   * The entity has already been built and internally validated, including adding
   * of children by Transaction::assemble. So convert any children to credcom
   * Entries and send them for validation.
   */
  static function transactionFormValidate($form, $form_state) {
    // $transaction = $form_state->getFormObject()->buildEntity($form, $form_state);
    // Would be good form to validate the transaction again,
    // and handle all the violations, again.
  }

  /**
   * Get the last hash which should match the trunkward node's.
   */
  static function getLastHash() {
    return \Drupal::database()
      ->select('mcapi_credcom_hashchain', 'cch')
      ->fields('cch', ['hash'])
      ->orderBy('id', 'desc')
      ->range(0, 1)
      ->execute()
      ->fetchField();
  }

  /**
   * @param Transaction $transaction
   * @return \CreditCommons\Entry
   */
  private static function ConvertTransactionToCcEntry(Transaction $transaction) : \CreditCommons\Entry {
    $data = (object)[
      'payer' => static::convertWalletToCredcomAcc($transaction->payer->entity),
      'payee' => static::convertWalletToCredcomAcc($transaction->payee->entity),
      'description' => $transaction->description->value,
      // Entry quants are always trunkward quants.
      'quant' => $transaction->worth->value * \Drupal::config('mcapi_credcom.settings')->get('rate')
    ];
    return \CreditCommons\Entry::create($data);
  }

  /**
   * @param array $values
   * @return \CreditCommons\Entry
   */
  private static function ConvertTransactionValuesToCcEntry(array $values) : \CreditCommons\Entry {
    $settings = \Drupal::config('mcapi_credcom.settings');
    $entry = [
      'description' => $values['description'][0]['value'],
      'quant' => $settings->get('rate') * $values['worth'][0]['value'], // all quants are trunkward_quants
    ];
    if (isset($values['payer'][0]['value'])) {
      $payee_wallet = Wallet::load($values['payee'][0]['target_id']);
      $entry['payee'] = $settings->get('cc_user') .'/'. str_replace(' ', '-', $payee_wallet->label());
      $entry['payer'] = $values['payer'][0]['value'];
    }
    else {
      $payer_wallet = Wallet::load($values['payer'][0]['target_id']);
      $entry['payer'] = $settings->get('cc_user') .'/'. str_replace(' ', '-', $payer_wallet->label());
      $entry['payee'] = $values['payee'][0]['value'];
    }
    return static::upcastEntry((object)$entry);
  }

  /**
   * @param Wallet $wallet
   * @return string
   */
  private static function convertWalletToCredcomAcc(Wallet $wallet) : \CreditCommons\Account {
    if ($wallet->id() == \Drupal::config('mcapi_credcom.settings')->get('wallet_id')) {
      return new AccountRemote (id: static::$remoteAccPath);
    }
    else {
      return new Account(id: str_replace(' ', '-', $wallet->name->value));
    }
  }

  /**
   * Utility, check if a credcom address refers to a local account, and if so
   * return the localised account name.
   *
   * @param string $cc_path
   * @return string
   *
   */
  static function credcomPathIsLocal(string $cc_path) : string {
    $conf = \Drupal::config('mcapi_credcom.settings');
    $local_name = '';
    $parts = explode('/', $cc_path);
    $pos = array_search($conf->get('node_name'), $parts);
    if ($pos !== FALSE) {
      $parts = array_slice($parts, $pos+1);
    }
    if (count($parts) == 1) {// Must be the intertrading wallet
      $first_part = reset($parts);
      if ($conf->get('trunkward_node_id') <> $first_part) {
        $local_name = str_replace('-', ' ', $first_part);
      }
    }
    return $local_name;
  }

  /**
   * @param \stdClass $data
   * @return type
   */
  static function upcastEntry(\stdClass $data) {
    foreach (['payee', 'payer'] as $role) {
      $str = $data->{$role};
      if ($local_name = static::credcomPathIsLocal($str)) {
        $acc = new Account($local_name);
      }
      else {
        if (!isset(static::$remoteAccPath)) {
          static::$remoteAccPath = $str;
        }
        $acc = new AccountRemote($str);
      }
      $data->{$role} = $acc;
    }
    return \CreditCommons\Entry::create($data);
  }

  /**
   * Exception handler to return credit commons errors to the network
   * Consider replacing with a class that extends Drupal\Core\Exception\ExceptionHandler
   * @param \Throwable $ex
   */
  static function CredcomExceptionHandler(\Throwable $ex) {
    $err = CCError::convertException($ex);
    \Drupal::logger('mcapi_credcom')
      ->error('@message', ['@message' => Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
    header('Content-Type: application/json');
    header('HTTP/1.1: 500 Server error');
    print json_encode(['errors' => [$err]]);
    exit;
  }

  /**
   * Error handler to return credit commons errors to the network
   * @param \Throwable $ex
   */
  static function CredcomErrorHandler($severity, $message, $filename, $lineno) {
    $trace = array_slice(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS), 1);
    if ($severity > 8000) {// deprecation messages
      \Drupal::logger('mcapi_credcom')
      ->warning(
        'Deprecated: @message\n@break\n@trace',
        ['@message' => $message, '@break'=> "$filename:$lineno", '@trace' => Markup::Create('<pre>'.print_r($trace, 1).'</pre>')]
      );
    }
    else {
      $err = new CCFailure($message);
      $err->trace = $trace;
      $err->break = $filename . ':'. $lineno;
      \Drupal::logger('mcapi_credcom')
        ->error('@message', ['@message' =>  Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
      header('Content-Type: application/json');
      header('HTTP/1.1: 500 Server error');
      print json_encode(['errors' => [$err]], 500);
      exit;
    }
  }

}
