<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\Workflow as CCWorkflowInterface;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Entity\Entity\EntityFormMode;

/**
 * Service handling outgoing requests to the trunkward node.
 */
class TrunkwardNodeRequester extends \CreditCommons\NodeRequester {

  private $logger;
  private $config;

  /**
   * Constructor
   *
   * @param ConfigFactory $config_factory
   * @param LoggerChannel $logger_channel
   */
  public function __construct(ConfigFactory $config_factory, LoggerChannel $logger_channel) {
    $this->config = $config_factory->get('mcapi_credcom.settings');
    $this->logger = $logger_channel;
    if ($this->configured()) {
      // Initiate the service from the settings.
      parent::__construct(
        $this->config->get('trunkward_node_url'),
        $this->config->get('cc_user'),
        Functions::getlastHash()
      );
    }
    $this->catch = TRUE;
  }

  public function configured() : bool {
    return $this->config->get('cc_user') and $this->config->get('trunkward_node_url');
  }

  /**
   * Cache the workflows.
   *
   * @return array
   *
   * @note workflows should change very rarely and even then should be
   * deprecated long before being removed.
   */
  public function getWorkflows() : array {
    $drupal_cache = \Drupal::Cache();
    $cached = $drupal_cache->get('credcom_workflows');
    if (!$cached) {
      foreach (parent::getWorkflows() as $wf) {
        if ($wf->creation->by) { // Only 1stparty workflows.
          $workflows[$wf->id] = $wf;
          $this->checkFormMode($wf);
        }
      }
      // make sure there is a transaction form mode for each.
      $drupal_cache->set('credcom_workflows', $workflows);
    }
    else {
      $workflows = $cached->data;
    }
    return $workflows;
  }

  /**
   * {@inheritDoc}
   *
   * Theoretically this could be disintermediated by javascript calling the
   * trunkward node directly, but then the creds would have to be sent to client.
   */
  public function accountNameFilter(string $acc_path = '', $limit = 10) : array {
    $result = [];
    try {
      $result = parent::accountNameFilter($acc_path, $limit);
    }
    catch (\Throwable $ex) {
      \Drupal::logger('mcpai_credcom')->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->makeMessage(), '@fullerror'=> $e]
      );
    }
    return $result;
  }

  public function convertPrice(string $node_path) : \stdClass {

  }

  public function buildValidateRelayTransaction(CCWorkflowInterface $transaction) : array {
    // Don't remove this because the parent expects the transaction object to be built.
    // @todo The cc-php-lib doesn't have a non-abstract transaction class
    return parent::buildValidateRelayTransaction($transaction);
    // CC Errors are caught in calling function mcapi_credcom_transaction_form_validate().
  }

  public function transactionChangeState(string $uuid, string $target_state) : void {
    try {
      parent::transactionChangeState($uuid, $target_state);
    }
    catch (\Throwable $e) {

    }
  }


  private function checkFormMode($workflow) : void {
    $id = 'mcapi_transaction.remote_'.$workflow->id;
    if ($mode = EntityFormMode::load($id)) {
      return;
    }

    $label = t('Remote @name', ['@name' => $workflow->label]);
    $form_mode = EntityFormMode::create([
      'targetEntityType' => 'mcapi_transaction',
      'label' => $label,
      'id' => $id,
      'cache' => TRUE
    ]);
    if (isset($wf->creation->by['payer'])) {
      $firstparty = 'payer';
      $secondparty = 'payee';
      $verb = 'crediting';
      $template = t("{{ payer }}\r\nRemote account: {{ payee }}\r\nThey received {{ description }}\r\nI am billing them for {{ worth }}");
    }
    else {
      $firstparty = 'payee';
      $secondparty = 'payer';
      $verb = 'billing';
      $template = t("{{ payee }}\r\nRemote account: {{ payer }}\r\nI received {{ description }}\r\nI am crediting them for {{ worth }}");
    }

    $settings = [
      'title' => $label,
      'incoming' => TRUE,
      'path' => '/transact/remote/'.$workflow->id,
      'menu_title' => $workflow->label,
      'menu_weight'  => 5,
      'transaction_type'  => 'remote',
      'experience_twig'  => "{{ $firstparty }}\r\nRemote account: {{ $secondparty }}\r\nWhat was received {{ description }}\r\nI am $verb them for {{ worth }}",
      'experience_button'  => t('Send')
    ];
    foreach ($settings as $k => $v) {
      $form_mode->setThirdPartySetting('mcapi_forms', $k, $v);
    }
    $form_mode->save();
  }
}
