<?php

/**
 * @file
 * Helper functions.
 */
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Form\FormStateInterface;

/**
 * Implements hook_form_MCAPI_CURRENCY_EDIT_FORM_alter().
 *
 * Add the limits settings to the currency form.
 */
function _mcapi_limits_form_mcapi_currency_edit_form_alter(&$form, FormStateInterface $form_state) {
  $form['limits'] = [
    '#title' => t('Balance limits'),
    '#description' => t("Transactions will be rejected which take accounts beyond these limits."),
    '#type' => 'details',
    '#weight' => 20,
    '#collapsible' => FALSE,
    '#open' => TRUE,
  ];
  $plugin = mcapi_limits_form_state_load_plugin(
    $form_state->getFormObject()->getEntity(),
    $form_state->getValue('plugin')
  );

  $form['limits'] += [
    'plugin' => [
      '#title' => t('Plugin'),
      '#description' => t('How will the limits be determined?'),
      '#type' => 'select',
      '#options' => \Drupal::service('plugin.manager.mcapi_limits')->optionsList(),
      // 'none' is the default plugin name.
      '#default_value' => $plugin->id,
      '#weight' => -1,
      '#ajax' => [
        'callback' => 'mcapi_limits_plugin_settings_subform',
        'wrapper' => 'limits-plugin-settings',
        'effect' => 'fade',
      ],
    ],
  ];
  // Show the options form for the selected plugin.
  $form['limits']['plugin_settings'] = $plugin->commonConfigurationForm([], $form_state) + [
    $plugin->id => $plugin->buildConfigurationForm([], $form_state),
    '#prefix' => '<div id="limits-plugin-settings">',
    '#suffix' => '</div>',
    '#tree' => TRUE
  ];
  // This is the ajax wrapper.
  $form['actions']['submit']['#validate'][] = 'mcapi_limits_validate';
  $form['actions']['submit']['#submit'][] = 'mcapi_limits_submit';
}

/**
 * Get the limits plugin.
 *
 * @param CurrencyInterface $currency
 *   The currency.
 * @param string $plugin_id
 *
 * @return \Drupal\mcapi_limits\Plugin\McapiLimitsInterface
 *   The Limits object.
 */
function mcapi_limits_form_state_load_plugin(CurrencyInterface $currency, string $plugin_id = NULL) {
  $limitsManager = \Drupal::service('plugin.manager.mcapi_limits');
  if ($plugin_id) {
    $plugin = $limitsManager->createInstance($plugin_id, ['currency' => $currency]);
  }
  else {
    $plugin = $limitsManager->createInstanceCurrency($currency);
  }
  return $plugin;
}

/**
 * Ajax callback.
 */
function mcapi_limits_validate(&$form, FormStateInterface $form_state) {
  $currency = $form_state->getFormObject()->getEntity();
  mcapi_limits_form_state_load_plugin($currency, $form_state->getValue('plugin'))
    ->validateConfigurationForm($form, $form_state);
}

/**
 * Form submission callback on the submit button.
 *
 * Save the plugin settings as config.
 */
function mcapi_limits_submit(&$form, FormStateInterface $form_state) {
  $currency = $form_state->getFormObject()->getEntity();
  $plugin = mcapi_limits_form_state_load_plugin($currency, $form_state->getValue('plugin'));
  $plugin->submitConfigurationForm($form, $form_state);

  // Rendered wallets may be affected by changes in balance limits.
  Cache::invalidateTags(['mcapi_wallet_view']);

  $config = $plugin->getConfiguration();
  $plugin_config = $config[$plugin->id]??[];
  unset($config[$plugin->id]);
  foreach ($config as $key => $value) {
    $currency->setThirdPartySetting('mcapi_limits', $key, $value);
  }
  $currency->setThirdPartySetting('mcapi_limits', $plugin->id, $plugin_config);
  // This has to go after the plugin has set the values.
  $currency->setThirdPartySetting('mcapi_limits', 'plugin', $plugin->id);
  $currency->save();
}

/**
 * Ajax callback.
 */
function mcapi_limits_plugin_settings_subform($form, $form_state) {
  return $form['limits']['plugin_settings'];
}

