<?php

namespace Drupal\mcapi_limits\Plugin;
use Drupal\mcapi\Entity\Type;
use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Base class for Limits plugins.
 *
 * @todo implement with PluginWithFormsTrait. Inject Logger
 */
abstract class McapiLimitsBase implements McapiLimitsInterface {

  use \Drupal\Core\StringTranslation\StringTranslationTrait;

  public $id;

  /**
   *
   * @var Drupal\mcapi\Entity\Currency
   */
  public $currency;

  /**
   *
   * @var Drupal\Core\Session\AccountInterface
   */
  public $currentUser;

  /**
   * Stores the limits settings from the currency, for convenience.
   *
   * @var array
   */
  protected $configuration;

  /**
   * Constructor.
   *
   * @param array $settings
   *   The plugin's settings.
   * @param string $plugin_id
   *   The id of the plugin.
   * @param array $definition
   *   Definition of the plugin.
   */
  public function __construct(array $settings, $plugin_id, array $definition) {
    $this->currency = $settings['currency'];
    $this->id = $plugin_id;
    $this->setConfiguration($this->currency->getThirdPartySettings('mcapi_limits'));
    // Not sure how to inject this.
    $this->currentUser = \Drupal::currentUser();
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $site_name = \Drupal::config('system.site')->get('name');
    return [
      'override' => 0,
      'skip' => ['inherit', 'user1']
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function commonConfigurationForm(array $form, FormStateInterface $form_state) {
    // We rely in the inserted fields to validate themselves individually, so
    // there is no validation added at the form level.
    $subform['override'] = [
      '#title' => $this->t('Allow wallet-level override'),
      '#description' => $this->t('Settings on the user profiles override these general limits.'),
      '#type' => 'checkbox',
      '#default_value' => $this->configuration['override'],
      '#weight' => 5,
      '#states' => [
        'invisible' => [
          ':input[name="limits[limits_callback]"]' => ['value' => 'limits_none'],
        ],
      ],
    ];
    $subform['skip'] = [
      '#title' => $this->t('Skip balance limit check for the following transactions'),
      '#description' => $this->t('Especially useful for mass transactions and automated transactions'),
      '#type' => 'checkboxes',
      // Casting it here saves us worrying about the default, which is awkward.
      '#default_value' => array_keys(array_filter($this->configuration['skip'])),
      // Would be nice if this was pluggable, but not needed for now.
      '#options' => [
        'auto' => $this->t("Automated transactions (such as automated fees)"),
        'user1' => $this->t("created by user 1"),
        'owner' => $this->t("created by the currency owner"),
      ],
      '#states' => [
        'invisible' => [
          ':input[name="limits[limits_callback]"]' => ['value' => 'limits_none'],
        ],
      ],
      '#weight' => 6,
    ];
    // To move this to the mcapi_mass module we would have to ensure the correct weight of modules
    if (Type::load('mass')) {
      $subform['skip']['#options']['mass'] = $this->t("Mass transactions"); // NB mass transactions are now a separate module
    }
    return $subform;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['empty'] = [
      '#markup' => $this->t('There are no settings for this plugin.')
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->hasValue('min') and $form_state->hasValue('max')) {
      if ($form_state->getValue('min') >= $form_state->getValue('max')) {
        $form_state->setError($form['min'], t('Min must be less than max.'));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $config = [];
    if ($form_state->getValue('plugin') != 'none') {
      $plugin_settings = $form_state->getValue(['plugin_settings', $this->id]);
      // Convert the worth field values.
      foreach ($this->getWorthFields() as $subfield) {
        $plugin_settings[$subfield] = $plugin_settings[$subfield]['value'] ?? NULL;
      }
      foreach ($plugin_settings as $key => $value) {
        $config[$this->id][$key] = $value;
      }
    }
    $this->setConfiguration($config);
  }

  /**
   * {@inheritdoc}
   */
  public function skipChecks(TransactionInterface $transaction) : bool {
    $config = $this->getConfiguration()['skip'];
    $replacements = ['@currency' => $this->currency->name];
    $logger = \Drupal::logger('mcapi');
    if (!empty($config['user1']) && $this->currentUser->id() == 1) {
      $logger->notice(
        'Skipped @currency balance limit check because you are user 1.',
        $replacements
      );
      return TRUE;
    }
    elseif (!empty($config['owner']) && $this->currentUser->id() == $this->currency->uid) {
      $logger->notice(
        'Skipped @currency balance limit check because you are the currency owner.',
        $replacements
      );
      return TRUE;
    }
    elseif (!empty($config['auto']) && $transaction->type->target_id == 'auto') {
      $logger->notice(
        'Skipped balance limit checks for @currency.',
        $replacements
      );
      return TRUE;
    }
    // @todo put this in the mass transaction module?
    elseif (!empty($config['mass']) && $transaction->type->target_id == 'mass') {
      $logger->notice(
        'Skipped balance limit checks for @currency.',
        $replacements
      );
      return TRUE;
    }
    return FALSE;
  }

  protected function getWorthFields() {
    return [];
  }

}
