<?php

namespace Drupal\mcapi_signatures\Form;

use Drupal\mcapi\Entity\Type;
use Drupal\mcapi\Plugin\TransactionRelativeManager;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for signature settings.
 */
class Settings extends ConfigFormBase {

  private $transactionRelativeManager;
  private $settings;

  /**
   * @param ConfigFactoryInterface $config_factory
   * @param TransactionRelativeManager $transactionRelativeManager
   */
  public function __construct(ConfigFactoryInterface $config_factory, TransactionRelativeManager $transactionRelativeManager) {
    parent::__construct($config_factory);
    $this->settings = $this->config('mcapi_signatures.settings');
    $this->transactionRelativeManager = $transactionRelativeManager;
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('mcapi.transaction_relative_manager')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mcapi_signatures_settings_form';
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // @todo This would look really tidy in a grid - but forms in tables are tricky
    $options = $this->transactionRelativeManager->options();
    unset($options['signatory'],
      $options['pending_signatory'],
      $options['creator'] // this would cause a paradox
    );
    $i= 0;
    $form['intro']['#markup'] = t("For each transaction type, specify which of the transactions relatives must to sign it before the transaction moves to workflow state 'completed'") . '. '.
      t("If a type has so signatories, it will be created in state 'completed'.");

    foreach (Type::loadMultiple() as $type) {
      if ($type->id == 'inherit') continue;
      $relations = $type->getThirdPartySetting('mcapi_signatures', 'relations', []);
      $form[$type->id()] = [
        '#title' => $this->t("Signatories for '@type' transactions", ['@type' => $type->label]),
        '#description' => $type->description,
        '#type' => 'details',
        '#collapsible' => TRUE,
        '#tree' => TRUE,
        '#open' => (bool)$relations,
        '#weight' => $i++,
        'relations' => [
          '#title' => t('Signatories'),
          '#type' => 'checkboxes',
          '#options' => $options,
          '#default_value' => $relations,
          '#required' => FALSE,
          '#weight' => 1,
        ],
        'notify' => [
          '#title' => t('Notify signatories by mail'),
          '#type' => 'checkbox',
          '#default_value' => in_array($type->id(), (array)$this->settings->get('types')),
          '#weight' => 2
        ]
      ];
    }

    // The mail template is REQURED if any of the type->notify flags is checked.
    $form['mail_template'] = [
      '#type' => 'fieldset',
      '#title' => t('Notification mail'),
      '#tree' => TRUE,
      'subject' => [
        '#title' => t('Notification subject'),
        '#type' => 'textfield',
        '#default_value' => $this->settings->get('mail_template.subject'),
        '#weight' => 1,
      ],
      'body' => [
        '#title' => t('Mail body'),
        '#description' => t("SHOULD include the token [xaction:sign-link]"),
        '#type' => 'textarea',
        '#weight' => $i++,
        '#default_value' => $this->settings->get('mail_template.body'),
        '#weight' => 2,
      ],
      '#weight' => $i++,
    ];
    if (\Drupal::moduleHandler()->moduleExists('token')) {
      $form['mail_template']['tokens'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => ['user', 'xaction', 'site', 'current-date'],
        '#global_types' => FALSE,
        '#weight' => 3,
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $mail_template = $form_state->getValue('mail_template');
    if (empty($mail_template['subject']) or empty($mail_template['body'])) {
      foreach (Type::loadMultiple() as $type) {
        if ($settings = $form_state->getValue($type->id)) {
          if ($settings['notify']) {
            $form_state->setErrorByName('mail_template', t('To notify signatories the mail template cannot be empty.'));
          }
        }
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $types = [];
    foreach (Type::loadMultiple() as $type) {
      if ($vals = $form_state->getValue($type->id())) {
        $type->setThirdPartySetting('mcapi_signatures', 'relations', array_filter($vals['relations']));
        $type->save();
        if ($vals['notify']) {
          $types[] = $type->id();
        }
      }
    }
    $this->settings
      ->set('mail_template', $form_state->getValue('mail_template'))
      ->set('types', $types)
      ->save();
    parent::submitForm($form, $form_state);
    $form_state->setRedirect('mcapi.admin');
  }

  /**
   * {@inheritDoc}
   */
  public function getEditableConfigNames() {
    return ['mcapi_signatures.settings'];
  }

}
