<?php

namespace Drupal\mcapi_signatures;

use Drupal\user\Entity\User;
use Drupal\mcapi\Event\McapiEvents;
use Drupal\mcapi\Event\TransactionSaveEvent;
use Drupal\mcapi\Event\TransactionAssembleEvent;
use Drupal\mailsystem\MailsystemManager;
use Drupal\Core\Logger\LoggerChannelFactory;
use Drupal\Core\Logger\LoggerChannel;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * A late subscriber to add signature slots to a transaction while building it.
 * @todo inject State
 */
class TransactionSubscriber implements EventSubscriberInterface {

  /**
   * @var Drupal\Core\Logger\LoggerChannel
   */
  private LoggerChannel $logger;

  /**
   * @var Drupal\mailsystem\MailsystemManager
   */
  private MailsystemManager $mailManager;

  function __construct(MailsystemManager $mail_manager, LoggerChannelFactory $logger_factory) {
    $this->mailManager = $mail_manager;
    $this->logger = $logger_factory->get('mcapi');
  }

  /**
   * {@inheritDoc}
   */
  public static function getSubscribedEvents() : array{
    return [
      // Must run AFTER any child transactions have been added.
      McapiEvents::ASSEMBLE => ['requireSignatures', -10],
      McapiEvents::ACTION => ['notifySignatories'],
    ];
  }

  /**
   * Add the requirement for signatures to a transaction during the assembly phase.
   *
   * @param TransactionAssembleEvent $event
   *   $event->getSubject() gives the transaction.
   */
  public function requireSignatures(TransactionAssembleEvent $event) {
    $transaction = $event->getSubject();
    if ($transaction->isNew() and !$transaction->isChild) {
      $t_type = $transaction->type->entity;
      if ($relatives = $t_type->getThirdPartySetting('mcapi_signatures', 'relations')) {
        // Signed by the current user
        mcapi_signature_service($transaction)->addSignatures()->sign();
        $transaction->state->setValue('pending');
      }
    }
  }

  /**
   * Send an email to the pending signatories.
   *
   * @param TransactionSaveEvent $event
   */
  public function notifySignatories(TransactionSaveEvent $event) {
    if (\Drupal::State()->get('mcapi_mail_stop')) {// used for transaction imports
      return;
    }
    $transaction = $event->getSubject();
    $args = $event->getArguments();
    if ($transaction->state->target_id == 'pending' and $args['old_state'] <> 'pending') {
      foreach (mcapi_signatures_notifications($transaction) as $uid) {
        $account = User::load($uid);
        $message = $this->mailManager->mail(
          'mcapi_signatures',
          'signhere',
          $account->getEmail(),
          $account->getPreferredLangcode(),
          ['user' => $account, 'xaction' => $transaction]
        );
        if ($message['result']) {
          $this->logger->notice(
            'Mailed %name to sign transaction @link',
            ['%name' => $account->getDisplayName(), '@link' => $transaction->toLink()->toString()]
          );
        }
      }
    }
  }

}
