<?php

namespace Drupal\mcapi\Entity;

use Drupal\Core\Render\Markup;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Defines the Currency entity.
 *
 * @ConfigEntityType(
 *   id = "mcapi_currency",
 *   label = @Translation("Currency"),
 *   label_collection = @Translation("Currencies"),
 *   handlers = {
 *     "access" = "Drupal\mcapi\Entity\Access\CurrencyAccessControlHandler",
 *     "view_builder" = "Drupal\mcapi\Entity\ViewBuilder\CurrencyViewBuilder",
 *     "form" = {
 *       "add" = "Drupal\mcapi\Form\CurrencyForm",
 *       "edit" = "Drupal\mcapi\Form\CurrencyForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm",
 *     },
 *     "list_builder" = "Drupal\mcapi\Entity\ListBuilder\CurrencyListBuilder",
 *     "route_provider" = {
 *       "html" = "Drupal\mcapi\Entity\RouteProvider\CurrencyRouteProvider",
 *     },
 *   },
 *   admin_permission = "configure mcapi",
 *   config_prefix = "currency",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "name",
 *     "uuid" = "uuid",
 *     "weight" = "weight",
 *     "owner" = "uid"
 *   },
 *   links = {
 *     "canonical" = "/currency/{mcapi_currency}",
 *     "collection" = "/admin/accounting/currencies",
 *     "edit-form" = "/admin/accounting/currencies/{mcapi_currency}",
 *     "add-form" = "/admin/accounting/currencies/add",
 *     "delete-form" = "/admin/accounting/currencies/{mcapi_currency}/delete",
 *   },
 *   config_export = {
 *     "id",
 *     "name",
 *     "provider",
 *     "uid",
 *     "description",
 *     "zero",
 *     "format",
 *     "weight"
 *   }
 * )
 * @todo EntityOwnerTrait applies only to ContentEntityTypes
 */
class Currency extends ConfigEntityBase implements CurrencyInterface {
  use StringTranslationTrait;

  /**
   * A short code identifying the currency within this site.
   *
   * @var string
   */
  public $id;

  /**
   * The UUID.
   *
   * @var string
   */
  public $uuid;

  /**
   * The name of the currency (plural)
   *
   * Plain text, probably with the first char capitalised.
   *
   * @var string
   */
  public $name;

  /**
   * Something about what the currency is for.
   *
   * @var string
   */
  public $description;

  /**
   * The owner user id.
   *
   * @var integer
   */
  public $uid;

  /**
   * The html string to be rendered if the transaction value is zero.
   *
   * An empty value means that zero transactions are not allowed.
   *
   * @var string
   */
  public $zero;

  /**
   * The weight of the currency when 'worth' is plural.
   *
   * @var integer
   */
  public $weight;

  /**
   * The format of the currency display.
   *
   * A keyed array defining how to convert the value from its raw value into a
   * formatted string or form element. Keys are 'f1', 'main', 'f2',
   * 'subdivision', 'f3.
   *
   * @var array
   */
  public $format;

  /**
   * {@inheritDoc}
   */
  public static function preCreate(EntityStorageInterface $storage_controller, array &$values) {
    $values += array(
      'format' => ['$', '999', '.', '99'],
      'zero' => FALSE,
      'weight' => 0,
      'uid' => \Drupal::currentUser()->id(),
    );
  }

  /**
   * {@inheritDoc}
   */
  public function transactionCount(array $conditions = [], $serial = FALSE) {
    $query = $this->entityTypeManager()
      ->getStorage('mcapi_transaction')->getQuery()->accessCheck(FALSE)
      ->count()
      ->condition('worth.curr_id', $this->id());
    foreach ($conditions as $field => $val) {
      $operator = is_array($val) ? 'IN' : '=';
      $query->condition($field, $val, $operator);
    }
    return $query->execute();
  }

  /**
   * {@inheritDoc}
   */
  public function volume(array $conditions = []) {
    return $this->entityTypeManager()
      ->getStorage('mcapi_transaction')
      ->volume($this->id(), $conditions);
  }

  /**
   * {@inheritDoc}
   */
  public function firstUsed() : int {
    return (int)$this->entityTypeManager()
      ->getStorage('mcapi_transaction')
      ->firstUsed($this);
  }

  /**
   * {@inheritDoc}
   */
  public function format($raw_num, $display = CurrencyInterface::DISPLAY_NORMAL, $linked = FALSE) {
    if ($display == CurrencyInterface::DISPLAY_NATIVE) {
      return $raw_num;
    }
    // If there is a minus sign this needs to go before everything, but
    // otherwise this function behaves the same.
    $minus_sign = $raw_num < 0 ? '-' : '';

    $parts = $this->formattedParts(abs($raw_num));
    // Now replace the parts back into the number format.
    $output = $this->format;
    foreach ($parts as $key => $num) {
      $output[$key] = $num;
    }
    if (isset($this->format['subdivision'])) {
      if ($pos = strpos($this->format['subdivision'], '/')) {
        // Don't use / in the pad length. ⅜
        $divisor = intval(substr($this->format['subdivision'], 0, $pos+1)) + 1;
        $pad_length = strlen(substr($this->format['subdivision'], 0, $pos));
        if ($display == CurrencyInterface::DISPLAY_NORMAL and $frac = $this->decToFraction($output['subdivision'] / $divisor)) {
          unset($output['f2']);
          $output['subdivision'] = $frac;
        }
        else {
          // no fraction found
          $output['subdivision'] = str_pad($output['subdivision'], $pad_length, '0', STR_PAD_LEFT);
        }
      }
      else {
        $pad_length = strlen($this->format['subdivision']);
        $output['subdivision'] = str_pad($output['subdivision'], $pad_length, '0', STR_PAD_LEFT);
      }
    }
    elseif ($pos = strpos($this->format['main'], '.')) {
      $dps = strlen($this->format['main']) - ($pos+1);
      $output['main'] = number_format((float)$output['main'], $dps, '.', '');
    }
    $output = $minus_sign . implode($output);
    // Try to display the value as a machine-readable number
    if ($display == CurrencyInterface::DISPLAY_PLAIN) {
      // Convert 1hr 23mins to 1.23 using regex
      // Replace likely separators with decimal point dots
      $output = preg_replace('/([.,:]+)/', '.', strip_tags($output));
      // Remove everything except numbers and dots and the minus sign.
      $output = preg_replace('/[^0-9.-]/', '', $output);
      return (float)$output;
    }
    $markup = Markup::create($output);
    // Link it to the currency stats page.
    if ($linked and $this->access('view')) {
      $markup = (string)$this->toLink($markup)->toString();
    }
    return $markup;
  }

  /**
   * {@inheritDoc}
   */
  function toLink($text = NULL, $rel = 'canonical', array $options = []) {
    $options = [
      'attributes' => [
        'title' => $this->t("View @currency dashboard", ['@currency' => $this->name]),
      ]
    ];
    return parent::toLink($text, $rel, $options);
  }

  /**
   * {@inheritDoc}
   */
  public function formattedParts(int $raw_num) : array {
    $formatted = [];
    if (empty($this->format['subdivision'])) {
      if ($pos = strrpos($this->format['main'], '.')) {
        $formatted['main'] = $raw_num * pow(10, -$pos);
      }
      else {
        $formatted['main'] = $raw_num;
      }
    }
    else {
      list($subs) = explode('/', $this->format['subdivision']);
      $divisor = (int)$subs + 1;
      if ($raw_num < 0) {
        $minus = TRUE;
        $raw_num = abs($raw_num);
      }
      $formatted['main'] = intval(floor($raw_num / $divisor));
      $formatted['subdivision'] = $raw_num % $divisor;
      if (isset($minus)) {
        $formatted['main'] = -$formatted['main'];
      }
    }
    return $formatted;
  }

  /**
   * Convert multiple components (coming from a form widget) into a raw value.
   *
   * @param array $parts
   *   Short array containing dollars & cents, or hours, minutes etc.
   *   EG [1=>9, 3=>99]
   *
   * @return int
   *   The raw value to be saved
   */
  public function unformat(array $parts) {
    $parts['main'] = $parts['main'] ? $parts['main'] + 0 : 0; //Make this numeric if it is null.
    if (empty($this->format['subdivision'])) {
      $raw_num = $parts['main'];
      if (strpos($this->format['main'], '.')) {
        $raw_num *= pow(10, strrpos($this->format['main'], '.'));
      }
    }
    else {
      list($atoms) = explode('/',  $this->format['subdivision']);
      if (!is_numeric($atoms)) {// could happen in migration. @todo in premigration check with a regex.
        trigger_error("problem converting currency format");
      }
      $divisor = (int)$atoms + 1;
      $fractions = is_numeric($parts['subdivision']) ? $parts['subdivision'] : 0;
      $raw_num = $parts['main'] * $divisor + $fractions;
    }
    return $raw_num;
  }

  /**
   * {@inheritDoc}
   */
  public function deletable() : bool {
    $all_transactions = $this->transactionCount();
    $deleted_transactions = $this->transactionCount(['state' => 'erased']);
    return $all_transactions == $deleted_transactions;
  }

  /**
   * {@inheritDoc}
   */
  public static function formats() {
    return [
      CurrencyInterface::DISPLAY_NORMAL => t('Normal'),
      CurrencyInterface::DISPLAY_NATIVE => t('Native integer'),
      CurrencyInterface::DISPLAY_PLAIN => t('Force decimal (Rarely needed)'),
    ];
  }

  /**
   * Generate a random value.
   *
   * The values should fall within a range 0-1 of the main counting unit i.e
   * dollars not cents, hours not minutes, and then something reasonably
   * rounded. We do this by choosing a nice number like 3.6  and unformatting
   * it.
   */
  public function sampleValue() {
    $parts['main'] = rand(2, 10);
    // Could be of the format 59/4.
    if (isset($this->format['subdivision'])) {
      $calc = explode('/', $this->format['subdivision']);
      $atoms = $calc[0] + 1;
      if (isset($calc[1])) {
        $parts['subdivision'] = rand(0, intval($calc[1])) * $atoms / $calc[1];
      }
      else {
        $parts['subdivision'] = rand(0, $num);
      }
    }
    return $this->unformat($parts);
  }

  /**
   * {@inheritDoc}
   *
   * Override to never invalidate the entity's cache tag; the config system
   * already invalidates it.
   */
  protected function invalidateTagsOnSave($update) {
    Cache::invalidateTags(['mcapi_currency_list']);
  }

  public function __toString() {
    return $this->name;
  }


  /**
   * {@inheritDoc}
   * Index currencies by key and sort.
   */
  static function loadMultiple(array $ids = NULL) {
    $currencies = parent::loadMultiple($ids);
    uasort($currencies, [get_class(), 'sort']);
    return $currencies;
  }

  /**
   * @param type $float
   * @return string
   * @kudos https://stackoverflow.com/questions/1954018/php-convert-decimal-into-fraction-and-back#9143510
   * @kudos https://www.alt-codes.net/fraction-symbols
   */
  function decToFraction($float) {
    // 1/2, 1/4, 1/8, 1/16, 1/3 ,2/3, 3/4, 3/8, 5/8, 7/8, 3/16, 5/16, 7/16,
    // 9/16, 11/16, 13/16, 15/16
    $whole = floor ( $float );
    $decimal = $float - $whole;
    $leastCommonDenom = 48; // 16 * 3;
    $denominators = array (2, 3, 4, 5);
    $roundedDecimal = round ( $decimal * $leastCommonDenom ) / $leastCommonDenom;
    if ($roundedDecimal == 0)
      return $whole;
    if ($roundedDecimal == 1)
      return $whole + 1;
    foreach ($denominators as $d ) {
      if ($roundedDecimal * $d == floor($roundedDecimal * $d )) {
        $denom = $d;
        break;
      }
    }
    $numerator = intval($roundedDecimal*$denom);
    if (isset($denom)) {
      switch ($denom) {
        case 2: //
          $frac = "&#189;"; #1/2
          break;
        case 3:
          if ($numerator == 1)$frac = '&#8531;'; # 1/3
          else $frac = '&#8532;'; # 2/3
          break;
        case 4:
          if ($numerator == 1)$frac = '&#188;'; # 1/4
          else $frac = '&#190;'; # 3/4
          break;
        case 5:
          if ($numerator == 1)$frac = '&#8533;';
          elseif ($numerator == 2)$frac = '&#8534;';
          elseif ($numerator == 3)$frac = '&#8535;';
          elseif ($numerator == 4)$frac = '&#8536;';
          break;
        default:
          //$frac = ($whole == 0 ? '' : $whole) . " " . ($roundedDecimal * $denom) . "/" . $denom;
      }
      return $frac;
    }

  }

}
