<?php

namespace Drupal\mcapi\Entity\Storage;

use Drupal\mcapi\Entity\Transaction;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Cache\Cache;

/**
 * Storage controller for Transaction entity.
 *
 * @todo inject ConfigFactory
 */
class TransactionStorage extends TransactionIndexStorage {

  /**
   * {@inheritDoc}
   *
   * Overriding this because transaction entity query should return serials keyed by xid
   */
  public function loadByProperties(array $values = array()) {
    $transactions = [];
    // Build a query to fetch the entity xids.
    $entity_query = $this->getQuery()->accessCheck(FALSE);
    if (!isset($values['xid'])) {
      // Don't attempt to load the children because there are in an entity reference field.
      $entity_query->noChildren();
    }
    $this->buildPropertyQuery($entity_query, $values);
    // Query returns serials keyed by xids.
    if ($serials = $entity_query->execute()) {
      $xids = array_keys($serials);
      $transactions = $this->loadMultiple($xids);
    }
    return $transactions;
  }

  /**
   * {@inheritDoc}
   *
   * Because the transaction entity is keyed by serial number not xid,
   * and because it contains child entities, we overwrite the whole save function.
   */
  public function save(EntityInterface $transaction) {
    if($transaction->state->target_id == 'temp') {
      // Ensure the transaction is not saved with temp state.
      // Other modules may change the state using hook_presave, hook_insert.
      $transaction->state->setValue('done');
    }
    $config = \Drupal::configFactory()->getEditable('mcapi.settings');
    $last = $config->get('last_serial');
    if ($transaction->serial->isEmpty()) {
      // this will set the serial for all the children.
      $transaction->serial->setValue($last + 1);
    }
    $config->set('last_serial', max($last, $transaction->serial->value))->save();
    $transaction->description->value = substr($transaction->description->value, 0, 255);
    return parent::save($transaction);
  }

  /**
   * {@inheritDoc}
   */
  public function resetCache(array $ids = NULL) {
    parent::resetCache($ids);
  }

  /**
   * {@inheritDoc}
   */
  function delete(array $entities) {
    // Delete all the child transactions as well
    foreach($entities as $transaction) {
      foreach ($transaction->children->referencedEntities() as $child) {
        $entities[$child->id()] = $child;
      }
    }
    parent::delete($entities);
  }

  /**
   * {@inheritDoc}
   */
  protected function doDelete($entities) {
    // First of all we need to get a flat array of all the entities.
    foreach ($entities as $xid => $transaction) {
      foreach ($transaction->flatten() as $tran) {
        $this->invalidateWalletCache($transaction);
        $entities[$xid] = $tran;
      }
    }
    parent::doDelete($entities);
  }

  /**
   * @param EntityInterface $entity
   * @param bool $update
   */
  protected function doPostSave(EntityInterface $entity, $update) {
    foreach ($entity->flatten() as $transaction) {
      $this->invalidateWalletCache($transaction);
    }
    parent::doPostSave($entity, $update);
  }

  /**
   * @param Transaction $transaction
   */
  private function invalidateWalletCache(Transaction $transaction) {
    $payer_tags = $payee_tags = [];
    // in cases of migration rollbacks the wallet may not be present
    if ($payer_wallet = $transaction->payer->entity) {
      $payer_tags = $payer_wallet->getCacheTagsToInvalidate();
    }
    if ($payee_wallet = $transaction->payee->entity) {
      $payee_tags = $payee_wallet->getCacheTagsToInvalidate();
    }
    Cache::invalidateTags(array_merge($payer_tags, $payee_tags));
  }

  /**
   * {@inheritDoc}
   */
  public function getQueryServiceName() {
    return 'mcapi_transaction.query.sql';
  }

}
