<?php

namespace Drupal\mcapi\Form;

use Drupal\user\Entity\User;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;

/**
 * Builder for currency entity form.
 */
class CurrencyForm extends EntityForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $currency = $this->entity;

    $form['name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name of currency'),
      '#description' => $this->t('Use the plural'),
      '#default_value' => $currency->label(),
      '#size' => 40,
      '#maxlength' => 80,
      '#required' => TRUE,
      '#weight' => 0,
    ];
    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description of the currency'),
      '#description' => $this->t('Why is this currency credible? Who guarantees it? What is its value? What is its story?'),
      '#default_value' => $currency->description,
      '#weight' => 1,
    ];

    if ($currency->id()) {
      $form['id'] = [
        '#type' => 'value',
        '#value' => $currency->id(),
      ];
    }
    else {
      $form['id'] = [
        '#type' => 'machine_name',
      // This is the max length of the worth->curr_id column.
        '#maxlength' => 8,
        '#machine_name' => [
          'exists' => '\Drupal\mcapi\Entity\Currency::load',
          'source' => ['name'],
        ],
        '#description' => $this->t('A unique machine-readable name for this Currency. It must only contain lowercase letters, numbers, and underscores.'),
      ];
    }
    if ($roles = user_roles(TRUE, 'manage mcapi')) {
      $admins = \Drupal::entityTypeManager()
        ->getStorage('user')
        ->loadByProperties(['roles' => array_keys($roles)]);
      if (count($admins)) {
        $options = [1 => User::load(1)->getDisplayName()];
        foreach ($admins as $uid => $user) {
          $options[$uid] = $user->getDisplayName();
        }
        $form['uid'] = [
          '#title' => $this->t('Comptroller'),
          '#description' => $this->t("Any user with 'Manage community accounting' role"),
          '#type' => 'select',
          '#options' => $options,
          '#default_value' => $currency->uid,
        ];
      }
    }
    if (!isset($form['uid'])) {
      $owner_id = $currency->uid ?? \Drupal::currentUser()->id();
      $form['uid'] = [
        '#type' => 'value',
        '#value' => $owner_id,
      ];
      $form['user_id'] = [
       '#type' => 'item',
       '#title' => $this->t('Comptroller'),
       '#markup' => User::load($owner_id)->toLink()->toString(),
       '#weight' => 3
      ];
    }
    $form['#attached']['library'][] = 'mcapi/currency-form';

    $serials = $currency->isNew() ?
      0 :
      $currency->transactionCount(['worth.value' => 0]);

    $form['zero'] = [
      '#title' => $this->t('Allow zero transactions'),
      '#description' => $this->t('Enter an HMTL snippet to show when a transaction is worth zero units. Leave blank to disallow zero transactions.'),
      '#type' => 'checkbox',
      '#default_value' => $currency->zero,
      // This is required if any existing transactions have zero value.
      '#weight' => 7,
    ];
    if ($serials) {
      $form['zero']['#required'] = TRUE;
      $form['zero']['#description'] .= $this->t("Zero transaction already exist so this field is required");
    }
    $help[] = $this->t('E.g. for Hours, Minutes put Hr99:59/4mins, for dollars and cents put $9.99; for loaves put 99 loaves.');
    $help[] = $this->t('The first number is always a string of zeros showing the number of characters (powers of ten) the widget will allow.');
    $help[] = $this->t('The optional /n at the end will render the final number widget as a dropdown field showing intervals, in the example, of 15 mins.');
    $help[] = $this->t('Special effects can be accomplised with HTML & css.');
    //Todo on rebuilding this form (for ajax) $currency->format is a string, as submitted
    $form['format'] = [
      '#title' => $this->t('Format'),
      '#description' => implode(' ', $help),
      '#type' => 'textfield',
      '#default_value' => is_array($currency->format) ? implode($currency->format) : $currency->format,
      '#element_validate' => [
        [$this, 'validateFormatElement'],
      ],
      '#max_length' => 32,
      '#size' => 32,
      '#weight' => 10
    ];
    $form['#attached']['library'][] = 'mcapi/currency-form';
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $currency = $this->entity;
    $status = $currency->save();

    if ($status == SAVED_UPDATED) {
      $this->messenger()->addStatus($this->t('Currency %label has been updated.', ['%label' => $currency->label()]));
    }
    else {
      $this->messenger()->addStatus($this->t('Currency %label has been added.', ['%label' => $currency->label()]));
    }
    $form_state->setRedirect('entity.mcapi_currency.collection');
  }

  /**
   * Element validation callback.
   */
  public function validateFormatElement(&$element, FormStateInterface $form_state) {
    try {
      $array = self::transformFormat($element['#value']);
      $form_state->setValue('format', $array);
    }
    catch (\Exception $e){
      $form_state->setError($element, $e->getMessage());
    }
  }

  private static function isFormatChar($char) {
    return is_numeric($char) or $char == '/' or $char == '.';
  }

  /**
   * Parse the format.
   * @param string $format
   * @return array
   */
  public static function transformFormat(string $format) : array {
    // Process in chunks because having a lot of trouble with long regexes
    $regexes = [
      'f1' => '[^9]*',
      'main' => '[9]+(\.(9+))?',
      'f2' => ' ?<.*>[^0-9]*|[^0-9]+', // Optional text, might include html tags like img or strong
      'subdivision' => '[0-9]+(\/[0-9]+)?', // Optional
      'f3' => '.*' // Optional
    ];
    $parts = [];
    $substr = $format;
    foreach ($regexes as $key => $pattern) {
      $result = preg_match("/$pattern/", $substr, $matches);
      // if there is a substring and no match
      if (empty($matches[0]) and $key <> 'f1') {
        throw new \Exception('Badly formatted');
      }
      $parts[$key] = $matches[0];
      if ($matches[0] == $substr) {
        break;
      }
      $substr = substr($substr, strlen($matches[0]));
    }
    if (count($parts) == 1) {
      throw new \Exception('Format must have text and numbers.');
    }
    return $parts;
  }


}
