<?php

namespace Drupal\mcapi\Form;

use Drupal\mcapi\Event\McapiEvents;
use Drupal\mcapi\Event\TransactionSaveEvent;
use Drupal\mcapi\TransactionOperations;
use Drupal\Core\Url;
use Drupal\Core\Render\Markup;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityConfirmFormBase;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Render\Renderer;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A form for confirming all operations on a transaction.
 */
class TransactionOperationForm extends EntityConfirmFormBase {

  /**
   * @var Drupal\system\Entity\Action;
   */
  private $plugin;

  /**
   * @var array
   */
  private $config;

  /**
   * @var ContainerAwareEventDispatcher
   */
  private $eventDispatcher;

  /**
   * @var Renderer
   */
  private $renderer;

  /**
   * @var Drupal\mcapi\Entity\ViewBuilder\TransactionViewBuilder
   */
  protected $transactionViewBuilder;

  /**
   * @param EntityTypeManager $entity_type_manager
   * @param ContainerAwareEventDispatcher $event_dispatcher
   * @param Renderer $renderer
   */
  public function __construct(EntityTypeManager $entity_type_manager, ContainerAwareEventDispatcher $event_dispatcher, Renderer $renderer) {
    $this->transactionViewBuilder = $entity_type_manager->getViewBuilder('mcapi_transaction');
    $this->eventDispatcher = $event_dispatcher;
    $this->renderer = $renderer;
    $op = $this->getRequest()->attributes->get('operation');
    $this->plugin = TransactionOperations::loadOperation($op);
    $this->config = $this->plugin->getPlugin()->getConfiguration();
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('event_dispatcher'),
      $container->get('renderer')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function buildform(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $form['#title'] = $this->getQuestion();
    $form['#attributes']['class'][] = 'transaction-operation-form';
    $form['description'] = $this->getDescription();

    $form['description']['#weight'] = -1;

    // By default, render the form using theme_confirm_form().
    $form += ['#theme' => 'confirm_form'];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function getDescription() {
    if ($this->config['format'] == 'twig') {
      $desc = [
        '#type' => 'inline_template',
        '#attached' => [
          'library' => ['mcapi/transaction']
        ],
        '#template' => $this->config['twig'],
        '#context' => ['mcapi_transaction' => $this->entity],
      ];
    }
    else {
      $desc = $this->transactionViewBuilder->view($this->entity, $this->config['format']);
    }
    return $desc;
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Use the entity as is, don't rebuild it
    $transaction = $this->entity;
    $transaction->setValidationRequired(FALSE);
    //$transaction->setChangedTime(\Drupal::time()->getRequestTime());
    try {
      // The op might have injected values into the form, so it needs to be able
      // to access them.
      $this->plugin->execute([$transaction]);
      if ($this->getOperation() == 'delete') {
        if (empty($this->destinationPath)) {
          // Front page.
          $this->destinationPath = '/';
        }
      }
      $event = new TransactionSaveEvent(
        $transaction,
        [
          'values' => $form_state->getValues(),
          'old_state' => $transaction->state->value,
          'action' => $this->getOperation(),
        ]
      );
      $this->eventDispatcher->dispatch($event, McapiEvents::ACTION);
      if ($m = $this->config['message']) {
        $this->messenger()->addMessage($m, 'status');
      }
    }
    catch (\Exception $e) {
      $this->messenger()->addError($this->t(
        "Error performing @action action: @error",
        [
          '@action' => $this->config['title'],
          '@error' => $e->getMessage(),
        ]
      ));
      return;
    }
    if (!empty($this->config['redirect'])) {
      $form_state->setRedirectUrl(Url::fromRoute($this->config['redirect']));
    }
    elseif (!$transaction->isNew()) {
      $redirect_url = $transaction->toUrl();
      $form_state->setRedirectUrl($transaction->toUrl());
    }
  }

  /**
   * {@inheritDoc}
   */
  public function getCancelUrl(): Url {
    if ($this->entity->isNew()) {
      $cancel_url = new Url('user.page');
    }
    else {
      $cancel_url = $this->entity->toUrl();
    }
    return $cancel_url;
  }

  /**
   * {@inheritDoc}
   */
  public function getCancelText() {
    return $this->config['cancel_link'];
  }

  /**
   * {@inheritDoc}
   */
  public function getConfirmText() {
    return $this->config['button'];
  }

  /**
   * {@inheritDoc}
   */
  public function getQuestion() {
    return Markup::create($this->config['title']);
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mcapi_transaction_' . $this->getOperation() . '_confirm_form';
  }

  /**
   * {@inheritDoc}
   */
  public function getBaseFormId() {
    return 'mcapi_transaction_confirm_form';
  }

  /**
   * {@inheritDoc}
   */
  protected function prepareEntity() {
    // Showing the form for the first time.
    // Load the transaction from the SerialConverter
    return $this->getRouteMatch()->getParameter('mcapi_transaction');
  }

  /**
   * {@inheritDoc}
   */
  public function buildEntity($element, $form_state) {
    // Called on form submission
    return $this->entity;
  }

}
