<?php

namespace Drupal\mcapi;

use Drupal\user\EntityOwnerInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Component\Utility\Html;

/**
 * Utility class for community accounting.
 */
class Mcapi {

  /**
   * Get a list of bundles which can have wallets.
   *
   * Which is to say, whether it is a contentEntity configured to have at least
   * 1 wallet.
   *
   * @return array
   *   Bundle names keyed by entity type ids.
   */
  public static function walletableBundles($reset = FALSE) : array {
    static $bundles = [];
    if (empty($bundles) || $reset) {
      // This is called during installation of config so holder isn't available yet.
      if ($fs = FieldStorageConfig::load('mcapi_wallet.holder')) {
        $fc = FieldConfig::load('mcapi_wallet.mcapi_wallet.holder');
        foreach ($fs->getSetting('entity_type_ids') as $entity_type) {
          $info = $fc->getSetting($entity_type);
          if (isset($info['handler_settings']['target_bundles'])) {
            $bundles[$entity_type] = $info['handler_settings']['target_bundles'];
          }
          else {
            $bundles[$entity_type][] = $entity_type;
          }
        }
      }
    }
    return $bundles;
  }

  /**
   * @param ContentEntityInterface $holder
   * @return @bool
   */
  static function holderIsWalletable(ContentEntityInterface $holder) : bool {
    $result = FALSE;
    if ($holder->getEntityTypeId() == 'user') {
      $result = TRUE;
    }
    elseif ($holder instanceOf EntityOwnerInterface) {
      $bundles = static::walletableBundles();
      $result = isset($bundles[$holder->getEntitytypeId()]) and in_array($holder->bundle(), $bundles[$holder->getEntitytypeId()]);
    }
    return $result;
  }

  /**
   * Get a list of entity labels keyed by entity ID suitable for a select box.
   * @param string $entity_type_id
   * @return string[]
   *
   * @note this is only used for mcapi_state, mcapi_type and mcapi_currency ATM
   */
  static function entityOptions(string $entity_type_id) : array {
    $entities = \Drupal::entityTypeManager()->getStorage($entity_type_id)->loadMultiple();
    $contextRepo = \Drupal::service('entity.repository');
    foreach ($entities as $entity) {
      $label = $contextRepo->getTranslationFromContext($entity)->label();
      $list[$entity->id()] = Html::escape($label);
    }
    return $list;
  }

  /**
   * Get a list of all the tokens on the transaction entity.
   *
   * @note This really needs replacing with a core function.
   */
  public static function tokenHelp() : string {
    foreach (array_keys(\Drupal::Token()->getInfo()['tokens']['xaction']) as $token) {
      $tokens[] = "[xaction:$token]";
    }
    return implode(', ', $tokens);
  }

  /**
   * Check the wallet holder field settings to see if only users can hold wallets.
   * @param bool $refresh
   * @return bool
   *
   * @todo move the state var to cache.
   */
  public static function usersOnly(bool $refresh = FALSE) : bool {
    static $users_only = NULL;
    $key = 'wallet_user_holders_only';
    if ($refresh or is_null($users_only)) {
      $state = \Drupal::state();
      if ($refresh or is_null($state->get($key))) {
        $entity_type_ids = FieldStorageConfig::load('mcapi_wallet.holder')
          ->getSetting('entity_type_ids');
        unset($entity_type_ids['user']);
        $users_only = empty($entity_type_ids);
        $state->set($key, $users_only);
      }
      else {
        $users_only = $state->get($key, FALSE);
      }
    }
    return $users_only;
  }

}
