<?php

namespace Drupal\mcapi\Plugin\Block;

use Drupal\mcapi\Entity\Wallet;
use Drupal\user\Entity\User;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Base block class for blocks determine a wallet from the context.
 */
abstract class WalletContextBlockBase extends BlockBase implements ContainerFactoryPluginInterface {

  const HOLDER_CURRENT_USER = 1;
  const HOLDER_VISITED_ENTITY = 0;

  /**
   * @var Drupal\Core\Session\AccountProxy
   */
  protected $entityTypeManager;

  /**
   * @var array
   */
  protected $wids = [];

  /**
   * @var Wallet[]
   */
  protected $wallets = [];

  /**
   * @var ContentEntity
   */
  protected $holder;

  /**
   * Constructor
   *
   * @param array $configuration
   * @param type $plugin_id
   * @param array $plugin_definition
   * @param EntityTypeManagerInterface $entity_type_manager
   * @param AccountInterface $current_user
   * @param Request $request
   */
  public function __construct($configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, AccountInterface $current_user, Request $request) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityTypeManager = $entity_type_manager;
    if ($this->configuration['holder_source'] == SELF::HOLDER_CURRENT_USER) {
      $this->holder = User::load($current_user->id());
    }
    elseif ($request->attributes and $request->attributes->has('_entity')) {
      $entity = $request->attributes->get('_entity');
      if ($entity instanceOf ContentEntityInterface) {
        $this->holder = $entity;
      }
    }
    if ($this->holder) {
      $this->wids = $entity_type_manager->getStorage('mcapi_wallet')->getQuery()
        ->accessCheck(TRUE)
        ->condition('holder.target_type', $this->holder->getEntityTypeId())
        ->condition('holder.target_id', $this->holder->id())
        ->execute();
    }
  }

  /**
   * Injection.
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('request_stack')->getCurrentRequest()
    );
  }

  /**
   * {@inheritDoc}
   */
  public function defaultConfiguration() {
    return [
      'holder_source' => SELF::HOLDER_CURRENT_USER,
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function blockAccess(AccountInterface $account) : AccessResult {
    if (empty($this->holder)) {
      $result = AccessResult::forbidden('No wallet holder on this route');
    }
    elseif (empty($this->wids)) {
      // Not sure how to cache this.
      $result = AccessResult::forbidden('Wallet holder has no wallets: '.$this->holder->getEntityTypeId() .' '. $this->holder->id());
    }
    else {
      $result = AccessResult::allowed()->cachePerUser();
    }
    return $result;
  }

  /**
   * {@inheritDoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $form['holder_source'] = [
      '#title' => $this->t('Entity source'),
      '#description' => $this->t('How to identify the entity whose wallet(s) we are displaying'),
      '#type' => 'radios',
      '#options' => array(
        static::HOLDER_VISITED_ENTITY => $this->t('From the route'),
        static::HOLDER_CURRENT_USER => $this->t('Logged in user'),
      ),
      '#default_value' => $this->configuration['holder_source'],
      '#weight' => 1
    ];
    return $form + parent::blockForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $form_state->cleanValues();
    parent::blockSubmit($form, $form_state);
    foreach ($form_state->getValues() as $key => $val) {
      $this->configuration[$key] = $val;
    }
  }

  /**
   * {@inheritDoc}
   * Build the block caching.
   */
  public function build() {
    $renderable = [];
    if ($this->configuration['holder_source'] == SELF::HOLDER_CURRENT_USER) {
      $renderable['#cache']['contexts'][] = 'user';
    }
    else {
      $renderable['#cache']['contexts'][] = 'url';
    }

    $renderable['#cache']['tags'] = array_map(
      function ($w){return 'mcapi_wallet:'.$w;},
      $this->wids
    );
    return $renderable;
  }

}
